@@ -1,4 +1 @@
-import doctest
-import moarchiving
-
 print('use ``python -m moarchiving.test`` to test the moarchiving module')
