@@ -1,24 +1,52 @@
 # -*- coding: utf-8 -*-
-"""This module contains, for the time being, a single MOO archive class.
-
-A bi-objective nondominated archive as sorted list with incremental
-update in logarithmic time.
+"""A bi-objective nondominated archive,
 
+implemented as sorted list and with incremental update in logarithmic time.
 """
 from __future__ import division, print_function, unicode_literals
-__author__ = "Nikolaus Hansen"
-__license__ = "BSD 3-clause"
-__version__ = "0.6.0"
-del division, print_function, unicode_literals
-
 import warnings as _warnings
 # from collections import deque  # does not support deletion of slices!?
 import bisect as _bisect # to find the insertion index efficiently
-try: import fractions
-except ImportError: _warnings.warn(
+try:
+    import fractions
+except ImportError:
+    _warnings.warn(
     '`fractions` module not installed, arbitrary precision hypervolume computation not available')
+del division, print_function, unicode_literals
+
 inf = float('inf')
 
+
+def _debug_trace(*args, **kwargs):
+    """return a string like printing the calling trace stack"""
+    try:
+        import traceback
+    except:
+        s = ''
+    else:
+        s = ''.join(traceback.format_stack(*args, **kwargs))
+    return s
+
+
+def true_fraction(val, copy=False):
+    """return a `fractions.Fraction` object from `val`.
+
+    Fixes the issue that `Fraction` does not convert an `np.intc` or
+    `np.int32` type to infinite representation `int`.
+    """
+    try:
+        fractions.Fraction
+    except NameError:
+        return val
+    if isinstance(val, fractions.Fraction):
+        if copy:  # Fraction(.) is almost 20 times slower than float(.)
+            return fractions.Fraction(val)
+        return val
+    if not isinstance(val, (int, float)):
+        val = float(val)
+    return fractions.Fraction(val)
+
+
 class BiobjectiveNondominatedSortedList(list):
     """A sorted list of non-dominated unique objective-pairs.
 
@@ -80,20 +108,34 @@ class BiobjectiveNondominatedSortedList(list):
     # to the hypervolume) are discarded. We may want to keep them, for simplicity
     # in a separate list?
 
-    # Default Values for respective instance attributes
+    # Default Values for meta control attributes
     make_expensive_asserts = False
-    try:
-        hypervolume_final_float_type = fractions.Fraction  # HV computation takes three times longer, precision may be more relevant here
-        hypervolume_computation_float_type = fractions.Fraction  # HV computation takes three times longer, precision may be less relevant here
-    except:
-        hypervolume_final_float_type = float  # lambda x: x is marginally faster
-        hypervolume_computation_float_type = float  # may be a good compromise
+
+    hypervolume_final_float_type = true_fraction
+    """HV computation takes increasingly longer with increasing precision (number of iterations).
+
+        Set ``BiobjectiveNondominatedSortedList.hypervolume_final_float_type = float``
+        when speed is an issue.
+        """ # lambda x: x is marginally faster than float
+    hypervolume_computation_float_type = true_fraction
+    """HV computation takes increasingly longer with increasing precision (number of iterations).
+
+        Precision may be less relevant here than for
+        `hypervolume_final_float_type`.
+
+        Set ``BiobjectiveNondominatedSortedList.hypervolume_computation_float_type = float``
+        here first when speed is an issue.
+        """
+
     maintain_contributing_hypervolumes = False
 
     def __init__(self,
                  list_of_f_pairs=None,
                  reference_point=None,
-                 sort=sorted):
+                 sort=sorted,
+                 infos=None,
+                 hypervolume_final_float_type=None,
+                 hypervolume_computation_float_type=None):
         """`list_of_f_pairs` does not need to be sorted.
 
         f-pairs beyond the `reference_point` are pruned away. The
@@ -105,10 +147,19 @@ class BiobjectiveNondominatedSortedList(list):
         CAVEAT: the interface, in particular the positional interface
         may change in future versions.
         """
+        if hypervolume_final_float_type is None:
+            self.hypervolume_final_float_type = BiobjectiveNondominatedSortedList.hypervolume_final_float_type
+        else:
+            self.hypervolume_final_float_type = hypervolume_final_float_type
+
+        if hypervolume_computation_float_type is None:
+            self.hypervolume_computation_float_type = BiobjectiveNondominatedSortedList.hypervolume_computation_float_type
+        else:
+            self.hypervolume_computation_float_type = hypervolume_computation_float_type
+
         self.make_expensive_asserts = BiobjectiveNondominatedSortedList.make_expensive_asserts
-        self.hypervolume_final_float_type = BiobjectiveNondominatedSortedList.hypervolume_final_float_type
-        self.hypervolume_computation_float_type = BiobjectiveNondominatedSortedList.hypervolume_computation_float_type
         self.maintain_contributing_hypervolumes = BiobjectiveNondominatedSortedList.maintain_contributing_hypervolumes
+        self.n_obj = 2
 
         if list_of_f_pairs is not None and len(list_of_f_pairs):
             try:
@@ -118,13 +169,30 @@ class BiobjectiveNondominatedSortedList(list):
             if len(list_of_f_pairs[0]) != 2:
                 raise ValueError("need elements of len 2, got %s"
                                  " as first element" % str(list_of_f_pairs[0]))
-            list.__init__(self, sort(list_of_f_pairs) if sort else list_of_f_pairs)
+            if sort is None:
+                list.__init__(self, list_of_f_pairs)
+            else:
+                if infos is not None:
+                    f_pair2info = dict(zip([tuple(f_pair) for f_pair in list_of_f_pairs], infos))
+                    list.__init__(self, sort(list_of_f_pairs))
+                    infos = [f_pair2info[tuple(f_pair)] for f_pair in self]
+                else:
+                    list.__init__(self, sort(list_of_f_pairs))
+
             # super(BiobjectiveNondominatedSortedList, self).__init__(sort(list_of_f_pairs))
         if reference_point is not None:
             self.reference_point = list(reference_point)
         else:
             self.reference_point = reference_point
-        self._infos = None
+
+        if infos is not None:
+            if len(infos) != len(list_of_f_pairs):
+                raise ValueError(f"need as many infos as f_pairs, got "
+                                 f"{len(infos)} infos and {len(list_of_f_pairs)} f_pairs")
+            self._infos = infos
+        else:
+            self._infos = None
+
         self.prune()  # remove dominated entries, uses in_domain, hence ref-point
         if self.maintain_contributing_hypervolumes:
             self._contributing_hypervolumes = self.contributing_hypervolumes
@@ -132,8 +200,35 @@ class BiobjectiveNondominatedSortedList(list):
         else:
             self._contributing_hypervolumes = []
         self._set_HV()
+
+        if reference_point is not None:
+            if self._hypervolume > 0:
+                self._hypervolume_plus = self._hypervolume
+            else:
+                if list_of_f_pairs is None or len(list_of_f_pairs) == 0:
+                    self._hypervolume_plus = -inf
+                else:
+                    self._hypervolume_plus = -min([self.distance_to_hypervolume_area(f)
+                                                   for f in list_of_f_pairs])
+        else:
+            self._hypervolume_plus = None
         self.make_expensive_asserts and self._asserts()
 
+    def _debug_info(self):
+        """return debug info as a list of (key, value) tuples"""
+        def cut_list(l, n=100):
+            n2 = int(n/2) - 2
+            try:
+                if len(l) > n:
+                    return l[:n2] + ['...'] + [l[int(len(l) / 2)]] + ['...'] + l[-n2:]
+            except:
+                pass
+            return l
+        return [('len', len(self)),
+                ('attributes', dict((k, cut_list(v)) for k, v in self.__dict__.items())),
+                ('self', cut_list(self)),
+               ]
+
     def add(self, f_pair, info=None):
         """insert `f_pair` in `self` if it is not (weakly) dominated.
 
@@ -172,6 +267,9 @@ class BiobjectiveNondominatedSortedList(list):
             raise ValueError("argument `f_pair` must be of length 2, was"
                              " ``%s``" % str(f_pair))
         if not self.in_domain(f_pair):
+            if self.hypervolume_plus is not None and self.hypervolume_plus < 0:
+                self._hypervolume_plus = max((self._hypervolume_plus,
+                                              -self.distance_to_hypervolume_area(f_pair)))
             self._removed = [f_pair]
             return None
         idx = self.bisect_left(f_pair)
@@ -214,7 +312,8 @@ class BiobjectiveNondominatedSortedList(list):
         if self._infos is not None:  # if the list exists it needs to be updated
             self._infos[idx] = info
         del self[idx + 1:idx2]  # can make `add` 20x faster
-        if self._infos: del self._infos[idx + 1:idx2]
+        if self._infos:
+            del self._infos[idx + 1:idx2]
         self._add_HV(idx)
         assert len(self) >= 1
         assert self._infos is None or len(self) == len(self.infos) == len(self._infos), (
@@ -236,6 +335,10 @@ class BiobjectiveNondominatedSortedList(list):
         >>> assert [2, 3] in nda and f_pair not in nda
         >>> if f_pair in nda:
         ...     nda.remove(f_pair)
+        >>> nda = BiobjectiveNondominatedSortedList()
+        >>> nda.add_list([[6, 6], [5, 7], [4, 8], [3, 9]])
+        >>> nda.remove(nda[-1])
+        >>> _ = nda.add([2, 10])
         >>> nda = BiobjectiveNondominatedSortedList._random_archive(p_ref_point=1)
         >>> for t in [None, float]:
         ...     if t:
@@ -255,11 +358,14 @@ class BiobjectiveNondominatedSortedList(list):
         """
         idx = self.index(f_pair)
         self._subtract_HV(idx)
+        if hasattr(self, '_hypervolume'):
+            self._hypervolume_plus = self._hypervolume if self._hypervolume > 0 else -inf
         self._removed = [self[idx]]
         del self[idx]  # == list.remove(self, f_pair)
-        if self._infos: del self._infos[idx]
+        if self._infos:
+            del self._infos[idx]
 
-    def add_list(self, list_of_f_pairs):
+    def add_list(self, list_of_f_pairs, infos=None):
         """insert a list of f-pairs which doesn't need to be sorted.
 
         This is just a shortcut for looping over `add`, but `discarded`
@@ -286,9 +392,11 @@ class BiobjectiveNondominatedSortedList(list):
         a small performance benefit.
         """
         removed = []
+        if infos is None:
+            infos = len(list_of_f_pairs) * [None]
         # should we better create a non-dominated list and do a merge?
-        for f_pair in list_of_f_pairs:
-            if self.add(f_pair) is not None:
+        for f_pair, info in zip(list_of_f_pairs, infos):
+            if self.add(f_pair, info=info) is not None:
                 removed += [self._removed]  # slightly faster than .extend
         self._removed = removed  # could contain elements of `list_of_f_pairs`
         self.make_expensive_asserts and self._asserts()
@@ -321,12 +429,11 @@ class BiobjectiveNondominatedSortedList(list):
 
     def copy(self):
         """return a "deep" copy of `self`"""
-        _warnings.warn('BiobjectiveNondominatedSortedList.copy has never been tested')
         nda = BiobjectiveNondominatedSortedList()
         for d in self.__dict__:
-            nda[d] = self[d]
+            setattr(nda, d, getattr(self, d))
         # now fix all mutable references as a true copy
-        list.__init__(nda, self)
+        list.__init__(nda, (p[:] for p in self))
         nda.reference_point = [xi for xi in self.reference_point]
         nda._hypervolume = self.hypervolume_final_float_type(self._hypervolume)  # with Fraction not necessary
         nda._contributing_hypervolumes = [hv for hv in self._contributing_hypervolumes]
@@ -343,6 +450,14 @@ class BiobjectiveNondominatedSortedList(list):
         Details: This method does a binary search in `self` using
         `bisect.bisect_left`.
         """
+        try:
+            return _bisect.bisect_left(self, f_pair, lowest_index)
+        except Exception:
+            pass
+        try:
+            f_pair = f_pair.tolist()
+        except Exception:
+            f_pair = list(f_pair)
         return _bisect.bisect_left(self, f_pair, lowest_index)
 
     def dominates(self, f_pair):
@@ -493,6 +608,43 @@ class BiobjectiveNondominatedSortedList(list):
             assert abs(self._hypervolume - self.compute_hypervolume(self.reference_point)) < 1e-12
         return self._hypervolume
 
+    @property
+    def hypervolume_plus(self):
+        """uncrowded hypervolume of the entire list w.r.t. the "initial" reference point.
+
+        `hypervolume_plus` equals to the hypervolume when the archive is
+        nonempty, otherwise it is the smallest Euclidean distance to the
+        hypervolume area (AKA reference domain) times -1 of any element
+        that was previously added but rejected because it did not dominate
+        the reference point.
+
+        Raise `ValueError` when no reference point was given initially.
+
+        Details: conceptually, the distance computation is based on the
+        nondominated archive as if it was not pruned by the reference
+        point.
+
+        >>> from moarchiving import BiobjectiveNondominatedSortedList as NDA
+        >>> a = NDA(reference_point=[1, 1])
+        >>> a.hypervolume_plus
+        -inf
+        >>> a.add([1, 2])
+        >>> a.hypervolume_plus
+        -1.0
+        >>> a.add([1, 1])
+        >>> a.hypervolume_plus
+        -0.0
+        >>> a.add([0.5, 0.5])
+        0
+        >>> float(a.hypervolume_plus)
+        0.25
+        """
+        if self.reference_point is None:
+            raise ValueError("to compute the hypervolume_plus a reference"
+                             " point is needed (must be given initially)")
+        return self._hypervolume_plus
+
+
     @property
     def contributing_hypervolumes(self):
         """`list` of contributing hypervolumes.
@@ -528,14 +680,17 @@ class BiobjectiveNondominatedSortedList(list):
 
         If `idx` is an `f_pair`, return contributing hypervolume of element
         with value `f_pair`. If `f_pair` is not in `self`, return
-        `hypervolume_improvement(f_pair)`.
+        `hypervolume_improvement(f_pair)`, i.e., its uncrowded contributing
+        hypervolume (which can be negative).
 
         The return type is ``self.hypervolume_computation_float_type` and
         by default `fractions.Fraction`, which can be converted to `float`
         like ``float(....contributing_hypervolume(idx))``.
         """
-        try: len(idx)
-        except TypeError: pass
+        try:
+            len(idx)
+        except TypeError:
+            pass
         else:  # idx is a pair
             if idx in self:
                 idx = self.index(idx)
@@ -563,8 +718,9 @@ class BiobjectiveNondominatedSortedList(list):
         unless the archive is empty and the point does not dominate the
         reference point.
 
-        Assumes that extreme points in the archive are in the reference
-        domain.
+        The implementation assumes that all points of the archive are in
+        the reference domain (and more extreme points have been pruned, as
+        it is the default behavior).
 
         Details: the distance for dominated points is computed by
         iterating over the relevant kink points ``(self[i+1][0],
@@ -628,8 +784,8 @@ class BiobjectiveNondominatedSortedList(list):
                 + max((0, f_pair[1] - self.reference_point[1]))**2)**0.5 \
                if self.reference_point else 0
 
-    def hypervolume_improvement(self, f_pair):
-        """return how much `f_pair` would improve the hypervolume.
+    def _hypervolume_improvement0(self, f_pair):
+        """deprecated and only used for testing: return how much `f_pair` would improve the hypervolume.
 
         If dominated, return the distance to the empirical pareto front
         multiplied by -1.
@@ -639,15 +795,19 @@ class BiobjectiveNondominatedSortedList(list):
         Overall this amounts to the uncrowded hypervolume improvement,
         see https://arxiv.org/abs/1904.08823
         """
+        save_infos = self._infos.copy() if self._infos is not None else None
+        save_hypervolume_plus = self._hypervolume_plus
         dist = self.distance_to_pareto_front(f_pair)
         if dist:
             return -dist
+        hv0 = self.hypervolume
         state = self._state()
         removed = self.discarded  # to get back previous state
         added = self.add(f_pair) is not None
         if added and self.discarded is not removed:
             add_back = self.discarded
-        else: add_back = []
+        else:
+            add_back = []
         assert len(add_back) + len(self) - added == state[0]
         hv1 = self.hypervolume
         if added:
@@ -658,8 +818,58 @@ class BiobjectiveNondominatedSortedList(list):
         if self.hypervolume_computation_float_type is not float and (
             self.hypervolume_final_float_type is not float):
             assert state == self._state()
+        if hv0 != self.hypervolume:
+            _warnings.warn("HV changed from %f to %f while computing hypervolume_improvement" %
+                           (hv0, self.hypervolume))
+        self._infos = save_infos
+        self._hypervolume_plus = save_hypervolume_plus
         return self.hypervolume_computation_float_type(hv1) - self.hypervolume
 
+    def hypervolume_improvement(self, f_pair):
+        """return how much `f_pair` would improve the hypervolume.
+
+        If dominated, return the distance to the empirical pareto front
+        multiplied by -1.
+        Else if not in domain, return distance to the reference point
+        dominating area times -1.
+
+        Overall this amounts to the uncrowded hypervolume improvement,
+        see https://arxiv.org/abs/1904.08823
+
+        Details: when ``self.reference_point is None`` and `f_pair` is
+        a new extreme point, the returned hypervolume improvement is
+        ``float('inf')``.
+
+        This method extracts a sublist first and thereby tries
+        to circumentvent to compute small differences between large
+        hypervolumes.
+        """
+        dist = self.distance_to_pareto_front(f_pair)
+        if dist:
+            return -dist
+        if self.reference_point is None:
+            if f_pair[0] < self[0][0] or f_pair[1] < self[-1][1]:
+                return inf
+        # find sublist that suffices to get the contributing volume
+        i0 = self.bisect_left(f_pair)
+        i1 = i0
+        while i1 < len(self) and f_pair[1] <= self[i1][1]:
+            # f_pair also dominates self[i1]
+            i1 += 1
+        r0 = self[i1][0] if i1 < len(self) else self.reference_point[0]
+        r1 = self[i0-1][1] if i0 > 0 else self.reference_point[1]
+        assaved = BiobjectiveNondominatedSortedList.make_expensive_asserts
+        BiobjectiveNondominatedSortedList.make_expensive_asserts = False  # prevent infinite recursion
+        sub = BiobjectiveNondominatedSortedList(self[i0:i1], reference_point=[r0, r1], sort=None)
+        BiobjectiveNondominatedSortedList.make_expensive_asserts = assaved
+        hv0 = sub.hypervolume
+        sub.add(f_pair)
+        res = self.hypervolume_computation_float_type(sub.hypervolume) - hv0
+        if BiobjectiveNondominatedSortedList.make_expensive_asserts:
+            assert abs(res - self._hypervolume_improvement0(f_pair)) < 1e-9 * (0.1 + res), (
+                        res, self._hypervolume_improvement0(f_pair))
+        return res
+
     def _set_HV(self):
         """set current hypervolume value using `self.reference_point`.
 
@@ -671,6 +881,8 @@ class BiobjectiveNondominatedSortedList(list):
         if self.reference_point is None:
             return None
         self._hypervolume = self.compute_hypervolume(self.reference_point)
+        if self._hypervolume > 0:
+            self._hypervolume_plus = self._hypervolume
         return self._hypervolume
 
     def compute_hypervolume(self, reference_point):
@@ -727,6 +939,11 @@ class BiobjectiveNondominatedSortedList(list):
             return None
         if idx1 is None:
             idx1 = idx0 + 1
+        if idx1 - idx0 == len(self):  # subtract HV of all points
+            assert idx0 == 0
+            dHV = -self._hypervolume
+            self._hypervolume *= 0  # keep type
+            return dHV
         if idx0 == 0:
             y = self.reference_point[1]
         else:
@@ -744,13 +961,21 @@ class BiobjectiveNondominatedSortedList(list):
         assert dHV <= 0  # and without loss of precision strictly smaller
         if ((Ff in (float, int) or isinstance(self._hypervolume, (float, int)))
                 and self._hypervolume != 0 and abs(dHV) / self._hypervolume < 1e-9):
-            _warnings.warn("_subtract_HV: %f + %f loses many digits of precision"
-                          % (dHV, self._hypervolume))
+            _warnings.warn("_subtract_HV: subtracting {:.16e} from {:.16e} loses many digits of precision"
+                          "\nSelf info: {}\nTraceback: {}".format(
+                               -dHV,
+                               self._hypervolume,
+                               self._debug_info(),
+                               _debug_trace()))
         self._hypervolume += Ff(dHV)
         if self._hypervolume < 0:
-            _warnings.warn("adding %.16e to the hypervolume lead to a"
-                          " negative hypervolume value of %.16e" %
-                          (dHV, self._hypervolume))
+            _warnings.warn("subtracting {:.16e} from the hypervolume lead to a"
+                          " negative hypervolume value of {:.16e}"
+                          "\nSelf info: {}\nTraceback: {}".format(
+                               -dHV,
+                               self._hypervolume,
+                               self._debug_info(),
+                               _debug_trace()))
         # assert self._hypervolume >= 0
         return dHV
 
@@ -759,7 +984,6 @@ class BiobjectiveNondominatedSortedList(list):
 
         TODO: also update list of contributing hypervolumes in case.
         """
-        dHV = self.contributing_hypervolume(idx)
         if self.maintain_contributing_hypervolumes:
             """Exerimental code:
             self._contributing_hypervolumes.insert(idx, dHV)
@@ -772,13 +996,19 @@ class BiobjectiveNondominatedSortedList(list):
             raise NotImplementedError("update list of hypervolumes")
         if self.reference_point is None:
             return None
+        dHV = self.contributing_hypervolume(idx)
         Ff = self.hypervolume_final_float_type
         if self._hypervolume and (
                         Ff in (float, int) or isinstance(self._hypervolume, (float, int))) \
-                and dHV / self._hypervolume < 1e-9:
-            _warnings.warn("_subtract_HV: %f + %f loses many digits of precision"
-                          % (dHV, self._hypervolume))
+                and abs(dHV) / self._hypervolume < 1e-9:
+            _warnings.warn("_add_HV: adding {} to HV={} loses many digits of precision"
+                           "\nSelf info: {}\nTraceback: {}".format(
+                               dHV,
+                               self._hypervolume,
+                               self._debug_info(),
+                               _debug_trace()))
         self._hypervolume += Ff(dHV)
+        self._hypervolume_plus = self._hypervolume
         return dHV
 
     def prune(self):
@@ -804,7 +1034,8 @@ class BiobjectiveNondominatedSortedList(list):
             i += 1
         removed += self[0:i]
         del self[0:i]
-        if self._infos: del self._infos[0:i]
+        if self._infos:
+            del self._infos[0:i]
         i = 1
         while i < len(self):
             i0 = i
@@ -829,15 +1060,19 @@ class BiobjectiveNondominatedSortedList(list):
                         break
             removed += self[i0r:ir]
             del self[i0:i]
-            if self._infos: del self._infos[i0:i]
+            if self._infos:
+                del self._infos[i0:i]
             i = i0 + 1
         self._removed = removed  # [p for p in removed if p not in self]
         if self.maintain_contributing_hypervolumes:
             # Old or experimental code:
             """
-            self._contributing_hypervolumes = [  # simple solution
-                self.contributing_hypervolume(i)
-                for i in range(len(self))]
+            ::
+
+                self._contributing_hypervolumes = [  # simple solution
+                    self.contributing_hypervolume(i)
+                    for i in range(len(self))]
+
             """
             raise NotImplementedError
         return nb - len(self)
@@ -847,18 +1082,16 @@ class BiobjectiveNondominatedSortedList(list):
         """`list` of f-pairs discarded in the last relevant method call.
 
         Methods covered are `__init__`, `prune`, `add`, and `add_list`.
-        Removed duplicates are not element of the discarded list.
-        When not inserted and not already in `self` also the input
-        argument(s) show(s) up in `discarded`.
+        Removed duplicates are not element of the discarded list except with
+        `__init__`. When not inserted and not already in `self` also the
+        input argument(s) show(s) up in `discarded`.
 
         Example to create a list of rank-k-non-dominated fronts:
 
         >>> from moarchiving import BiobjectiveNondominatedSortedList as NDA
         >>> all_ = [[0.1, 1], [-2, 3], [-4, 5], [-4, 5], [-4, 4.9]]
-        >>> nda_list = [NDA(all_)]  # rank-0-non-dominated
-        >>> while nda_list[-1].discarded:
-        ...     nda_list += [NDA(nda_list[-1].discarded)]
-        >>> assert [len(p) for p in nda_list] == [3, 1]
+        >>> nda_list = NDA(all_)  # rank-0-non-dominated
+        >>> assert nda_list.discarded == [[-4, 5], [-4, 5]]
 
         """
         try:
@@ -904,16 +1137,22 @@ class BiobjectiveNondominatedSortedList(list):
         ... except ValueError: pass
         ... else: raise AssertionError("remove did not raise ValueError")
 
-        >>> from numpy.random import rand
+        >>> import numpy as np
+        >>> randn = np.random.randn
         >>> for _ in range(120):
         ...     a = moarchiving.BiobjectiveNondominatedSortedList._random_archive()
         ...     a.make_expensive_asserts = True
         ...     if a.reference_point:
-        ...         for f_pair in rand(10, 2):
+        ...         for i, f_pair in enumerate([randn(2) + [i, -i] for i in range(10)] +
+        ...                                    [randn(2) / randn(2) + [i, -i] for i in range(10)]):
+        ...             if i % 4 == 1:
+        ...                 _ = a.add(f_pair)
         ...             h0 = a.hypervolume
         ...             hi = a.hypervolume_improvement(list(f_pair))
-        ...             assert a.hypervolume == h0  # works OK with Fraction
-
+        ...             hi_org = a._hypervolume_improvement0(list(f_pair))
+        ...             assert hi == hi_org  # didn't raise with rand instead of randn
+        ...             assert a.hypervolume == h0, (a.hypervolume, h0)  # works OK with Fraction
+        ...             a._asserts()
 
         """
         assert sorted(self) == self
@@ -956,6 +1195,7 @@ class BiobjectiveNondominatedSortedList(list):
                 assert all(diffs[:, 0] > 0)
                 assert all(diffs[:, 1] < 0)
 
+
 if __name__ == "__main__":
     import doctest
     print('doctest.testmod() in moarchiving.py')
